package milab.lab02.BookSellerAgent;

import jade.core.Agent;
import jade.core.behaviours.*;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;

import javax.xml.bind.*;

import milab.lab02.BookSellerAgent.catalog.*;

import java.util.*;
import java.io.File;

import milab.*;

@SuppressWarnings("serial")
public class BookSellerAgent extends Agent {

	// The catalogue of books for sale
	private CatalogType catalog;
	// The GUI by means of which the user can add books in the catalogue
	private BookSellerGui myGui;
	// The filePath of the XML database holding the catalogue of books to sell
	private String xmlDB;
	// The filePath of the XSD schema describing the structure of the xmldb database
	private String xmlDBSchema = "milab/lab02/BookSellerAgent/catalog/catalog.xsd";
	// The package where the - JAXB generated - Java Context of the dbschema is
	private String xmlDBSchemaContext = "milab.lab02.BookSellerAgent.catalog";

	// Put agent initializations here
	protected void setup() {

		// Create and show the GUI 
		myGui = new BookSellerGui(this);
		myGui.show();

		// Register the book-selling service in the yellow pages
		DFAgentDescription dfd = new DFAgentDescription();
		dfd.setName(getAID());
		ServiceDescription sd = new ServiceDescription();
		sd.setType("book-selling");
		sd.setName("JADE-book-trading");
		dfd.addServices(sd);

		try {

			DFService.register(this, dfd);

		} catch (FIPAException fe) {

			fe.printStackTrace();

		}

		// Get the catalogue database's filePath as a start-up argument
		Object[] args = getArguments();

		if (args != null && args.length > 0) {

			xmlDB = (String) args[0];
			File xmlDatabase = new File(xmlDB);

			if (xmlDatabase.isFile()) {

				System.out.println("Catalogue database of " + getLocalName()
						+ " is " + xmlDB);

				// Create the catalog
				JAXBUnMarshaller jaxbUnMarshaller = new JAXBUnMarshaller();
				JAXBElement<CatalogType> catalogElement = (JAXBElement<CatalogType>) jaxbUnMarshaller
						.readXMLDocument(xmlDB, xmlDBSchema, xmlDBSchemaContext);
				catalog = catalogElement.getValue();

				// Add the behaviour serving queries from buyer agents
				addBehaviour(new OfferRequestsServer());

				// Add the behaviour serving purchase orders from buyer agents
				addBehaviour(new PurchaseOrdersServer());

			} else {

				// Make the agent terminate
				System.out
						.println("The catalogue database specified is non existing. Terminating...");
				doDelete();

			}

		} else {

			// Make the agent terminate
			System.out
					.println("No catalogue database specified. Terminating...");
			doDelete();

		}

	}

	// Put agent clean-up operations here
	protected void takeDown() {

		// Deregister from the yellow pages
		try {

			DFService.deregister(this);

		} catch (FIPAException fe) {

			fe.printStackTrace();

		}

		// Close the GUI
		myGui.dispose();
		// Printout a dismissal message
		System.out.println("Seller-agent " + getAID().getName()
				+ " terminating.");

	}

	/*
	   This is invoked by the GUI when the user adds a new book for sale (that is why it adds a behaviour: because it'll run in the agent's own thread unlike the GUI, which's events are executed in the eventDispatcher's thread...)
	 */
	public void updateCatalog(final String title, final int price) {

		addBehaviour(new OneShotBehaviour() {

			public void action() {

				// Creating the new book object to be inserted into the database
				List<BookType> bookList = catalog.getBook();
				BookType newbook = new BookType();
				newbook.setTitle(title);
				newbook.setPrice(price);
				newbook.setRating("100");

				// Addig the new book to the database (but first to the catalog)
				bookList.add(newbook);
				milab.lab02.BookSellerAgent.catalog.ObjectFactory factory = new milab.lab02.BookSellerAgent.catalog.ObjectFactory();
				JAXBElement<CatalogType> catalogElement = factory
						.createCatalog(catalog);
				JAXBMarshaller jaxbMarshaller = new JAXBMarshaller();
				jaxbMarshaller.generateXMLDocument(xmlDB, catalogElement, xmlDBSchemaContext);
				System.out.println(title + " inserted into the catalogue of "
						+ getLocalName() + ". Price = " + price);

			}

		});

	}

	/**
	 * Search the catalog using title a keyword string patterns.
	 * Available methods: title, keyword or both (AND).
	 * @param titlePattern the title must contain this string
	 * @param keywordPattern this keyword must present
	 * @param ontology check the keywords in this ontology 
	 * @return list of books
	 */
	public List<BookType> searchCatalog(String titlePattern,
			String keywordPattern, String ontology) {

		List<BookType> ret = new ArrayList<BookType>();

		// a list to search in the catalog
		List<BookType> searchList = catalog.getBook();

		// title search?
		if (titlePattern != null && titlePattern.length() > 0) {

			// Searching the list of books for a book with the given title
			for (int i = 0; i < searchList.size(); i++) {
				BookType book = (BookType) searchList.get(i);
				if (titlePattern.equalsIgnoreCase(book.getTitle())) {
					if (!ret.contains(book))
						ret.add(book);
				}
			}

			// Searching the list of books for a book with similar title
			for (int i = 0; i < searchList.size(); i++) {
				BookType book = (BookType) searchList.get(i);
				if (book.getTitle().contains(titlePattern)) {
					if (!ret.contains(book))
						ret.add(book);
				}
			}
		}

		/*
		 * uncomment the next part to activate keyword search
		 * 
		 */
		if (keywordPattern != null && keywordPattern.length() > 0) {

			// search in the previous (title search) results if there are any
			if (ret.size() > 0) {
				searchList = ret;
				ret = new ArrayList<BookType>();
			}

			// Searching the list of books for a book with a given keyword
			for (int i = 0; i < searchList.size(); i++) {
				BookType book = (BookType) searchList.get(i);
				List<BookType.Keywords> keywords_array = book.getKeywords();
				for (int j = 0; j < keywords_array.size(); j++) {
					BookType.Keywords keywords = keywords_array.get(j);

					// TODO check the book and search ontologies and skip this book if they differ

					
					boolean isExpanded = false;
					// try to find related keywords using an ontology
					/* TODO query expansion using an ontology
					// determining the query ontology
					SimpleOntology onto=null;
					try {
						Class o = Class.forName(ontology);
						onto = (SimpleOntology) o.newInstance();
						// do the expansion
						isExpanded = true;
					} catch (Exception e) {
						// ontology class not found - ignore the ontology search
						// isExpanded remains false
					}
					 */

					List<String> keywordList = keywords.getKeyword();
					for (int k = 0; k < keywordList.size(); k++) {
						String keyword = keywordList.get(k);
						if (keyword.equalsIgnoreCase(keywordPattern)) {
							if (!ret.contains(book)) {
								ret.add(book);
							}
							break; // book added - break out of the cycle
						}
						/* TODO query expansion using the ontology 
						if (isExpanded && ...) {
							if (!ret.contains(book)) {
								ret.add(book);
							}
							break;	// book added - break out of the cycle
						}
						 */
					}
					// do not continue the cycle if the book is already added to the list of found books
					if (ret.contains(book))
						break;
				}
			}
		} /* end of keyword search */

		return ret;

	}

	/*
	 * This is invoked by the agent(thread) itself (so it mustn't add a new
	 * behaviour...)
	 */
	public BookType removeCatalog(String title) {

		BookType ret = null;

		// Creating the list of books available in the catalogue
		List<BookType> bookList = catalog.getBook();
		// Searching the list of books for a book with the given title
		for (int i = 0; i < bookList.size(); i++) {

			BookType book = (BookType) bookList.get(i);
			if (title.equalsIgnoreCase(book.getTitle())) {

				ret = book; // Let's return the price of the book (as an Integer)

				// Removing book from the database (but first from the catalog)
				bookList.remove(i);
				milab.lab02.BookSellerAgent.catalog.ObjectFactory factory = new milab.lab02.BookSellerAgent.catalog.ObjectFactory();
				JAXBElement<CatalogType> catalogElement = factory.createCatalog(catalog);
				JAXBMarshaller jaxbMarshaller = new JAXBMarshaller();
				jaxbMarshaller.generateXMLDocument(xmlDB, catalogElement,xmlDBSchemaContext);
				break;

			}

		}

		return ret;

	}

	/*
	   Inner class OfferRequestsServer.
	   This is the behaviour used by Book-seller agents to serve incoming requests 
	   for offer from buyer agents.
	   If the requested book is in the local catalogue the seller agent replies 
	   with a PROPOSE message specifying the price. Otherwise a REFUSE message is
	   sent back.
	 */
	private class OfferRequestsServer extends CyclicBehaviour {

		@SuppressWarnings("unchecked")
		public void action() {

			MessageTemplate mt = MessageTemplate
					.MatchPerformative(ACLMessage.CFP);
			ACLMessage msg = myAgent.receive(mt);

			if (msg != null) {

				// CFP Message received. Process it
				JAXBUnMarshaller jaxbUnMarshaller = new JAXBUnMarshaller(); // Creating an UnMarshaller

				// TODO change this to use a search language
				JAXBElement<BookType> contentElement =
					(JAXBElement<BookType>) jaxbUnMarshaller.readXMLString(
								msg.getContent(),
								"milab/lab02/language/lang_books/lang_books.xsd",
								"milab.lab02.language.lang_books");
				// Taking the ACLMessageContent String as an XML, validating it, and unmarshalling it into an appropriate Java Object.
				BookType pattern = contentElement.getValue(); // Getting the String value of the Java Object representing the XML element...
				ACLMessage reply = msg.createReply();

				// TODO change this to use a search language
				List<BookType> foundBooks = searchCatalog(pattern.getTitle(), "", "");
				milab.lab02.language.lang_books.ObjectFactory factory = new milab.lab02.language.lang_books.ObjectFactory();

				// the content of the reply is a Book offer
				JAXBElement<BookType> replyContentElement;

				if (foundBooks.size() > 0) {
					// The requested book is available for sale. Reply with the price
					reply.setPerformative(ACLMessage.PROPOSE);
					// create the XML content element from the first book found. TODO all books?
					replyContentElement = factory.createBookContent(foundBooks
							.get(0));
				} else {
					// The requested book is NOT available for sale.
					reply.setPerformative(ACLMessage.REFUSE);
					BookType foundBook = new BookType();
					foundBook.setTitle("not-available");
					replyContentElement = factory.createBookContent(foundBook);
				}
				// Marshalling the Java Object into a String with appropriate XML syntax	        
				JAXBMarshaller jaxbMarshaller = new JAXBMarshaller(); // Creating a universal Marshaller
				reply.setContent(jaxbMarshaller.generateXMLString("UTF-8",
								replyContentElement,
								"milab.lab02.language.lang_books"));
				// Sending the reply to the buyer
				myAgent.send(reply);

			} else {

				block();

			}

		}

	} // End of inner class OfferRequestsServer

	/**
	   Inner class PurchaseOrdersServer.
	   This is the behaviour used by Book-seller agents to serve incoming 
	   offer acceptances (i.e. purchase orders) from buyer agents.
	   The seller agent removes the purchased book from its catalogue 
	   and replies with an INFORM message to notify the buyer that the
	   purchase has been sucesfully completed.
	 */
	private class PurchaseOrdersServer extends CyclicBehaviour {

		public void action() {

			MessageTemplate mt = MessageTemplate
					.MatchPerformative(ACLMessage.ACCEPT_PROPOSAL);
			ACLMessage msg = myAgent.receive(mt);

			if (msg != null) {

				// ACCEPT_PROPOSAL Message received. Process it
				JAXBUnMarshaller jaxbUnMarshaller = new JAXBUnMarshaller(); // Creating an UnMarshaller
				JAXBElement<BookType> contentElement =
					(JAXBElement<BookType>) jaxbUnMarshaller.readXMLString(
								msg.getContent(),
								"milab/lab02/language/lang_books/lang_books.xsd",
								"milab.lab02.language.lang_books");
				// Taking the ACLMessageContent String as an XML, validating it, and unmarshalling it into an appropriate Java Object.
				BookType soldBook = contentElement.getValue(); // Getting the String value of the Java Object representing the XML element...
				ACLMessage reply = msg.createReply();

				BookType removedBook = removeCatalog(soldBook.getTitle());
				if (removedBook != null) {

					reply.setPerformative(ACLMessage.INFORM);
					System.out.println(removedBook.getTitle()
							+ " sold to agent " + msg.getSender().getName());

				} else {

					// The requested book has been sold to another buyer in the meanwhile .
					reply.setPerformative(ACLMessage.FAILURE);

					milab.lab02.language.lang_books.ObjectFactory factory = new milab.lab02.language.lang_books.ObjectFactory();
					removedBook = new BookType();
					removedBook.setTitle("not-available");
					contentElement = factory.createBookContent(removedBook); // Creating the Java Object corresponding to the XML object
					JAXBMarshaller jaxbMarshaller = new JAXBMarshaller(); // Creating a universal Marshaller
					reply.setContent(jaxbMarshaller.generateXMLString("UTF-8",
							contentElement, "milab.lab02.language.lang_books")); // Marshalling the Java Object into a String with appropriate XML syntax	        

				}

				myAgent.send(reply);

			} else {

				block();

			}

		}

	} // End of inner class PurchaseOrdersServer

}
